#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>

#define _TCL_UTILS_H

#include "cl.h"
#include "IO.h"
#include "qual.h"
/*#include "gap_globals.h"*/
#include <dna_utils.h>
#include <tcl.h>

/* extern Tcl_Interp* our_interp; */

extern int consensus_mode;
#define CONSENSUS_MODE_FREQ       0
#define CONSENSUS_MODE_CONFIDENCE 2

/* Pinched from tagUtils.h */

#define type2str(t,s) \
( \
((s)[0] = ((t) >> 24) & 0xff), \
  ((s)[1] = ((t) >> 16) & 0xff), \
  ((s)[2] = ((t) >>  8) & 0xff), \
  ((s)[3] = ((t) >>  0) & 0xff), \
  ((s)[4] = '\0'), \
  (s))

typedef struct
{
  GCardinal read;
  long startInContig;
  long endInContig;
  long startInRead;
  long endInRead;
} AFInfo;

typedef struct
{
  char * memory;
  char * current;
  size_t alloc_length;
  size_t used;
  size_t size;
} GrowMem;

void initGrowMem(GrowMem * growable, size_t unitSize);
void gm_push(GrowMem * growable, void * object);
size_t size_of(GrowMem * groawable);
void lookup(GrowMem * growable, size_t item, void * object);
void freeGrowMem(GrowMem * growable);
void * mallocate(size_t bytes);
void * reallocate(void * old, size_t bytes);
void usage(void);
GapIO *open_database(char *project, char *version);
void gap2caf(FILE *fp, GapIO *dbhandle, int cutoff, int staden, int fasta,
	     int quality, int silent, int pads, int depad, int preserve,
	     char *contig_prefix, int bayes, int id);
void writeContig(FILE *fp, GapIO *dbhandle, int cutoff,	int staden, int fasta,
		 int quality, int silent, int pads, int depad, int preserve,
		 char *contig_prefix, int bayes, int contig,
		 char *used_annotations, char *used_notes, int id);
void writeRead(FILE *fp, GapIO *dbhandle, int quality, int silent,
	       int preserve, GCardinal reading, GReadings * readInfo,
	       GrowMem * assembledFrom, char *used_annotations,
               char *used_notes);
void writeDNA(FILE *fp, char *name, char *dna, int len, int fasta, int depad,
	      int pads);
void staden2caf(char *seq, int length);
void staden2fasta(char *seq, int length, int pads);
void writeQuality(FILE *fp, char *name, int1 *qual, int len);
void writeConsensusQuality(FILE *fp, char *name, float *qual, int len);
void writeAlignSCF(FILE *fp, char *name, int2 *op, int len);
void writeTraceFile(FILE *fp, GapIO *dbhandle, int reading);
void writeReadInfo(FILE *fp, GapIO *dbhandle, int reading,
		   GReadings * readInfo, int silent);
int fixTemplate(char *template);
int writeTags(FILE * fp, GapIO * dbhandle, GCardinal tag, char *used);
int writeNotes(FILE * fp, GapIO * dbhandle, GCardinal note, char *used);
void mapTypeToMethod(char t[5],
		     char *comment,
		     char **method,
		     char **newcomment);
int getTagValue (char *comment,
		 char *tag,
		 char **value,
		 char **newcomment,
		 int removeIfFound);
int find_end_of_value (char *comment, int i,char sep);
void excise(char *comment,
	    int from,
	    int len,
	    char **newcomment,
	    char sep);

void initGrowMem(GrowMem * growable, size_t unitSize)
{
  growable->memory = (char *) mallocate(16 * unitSize);
  growable->current = growable->memory;
  growable->alloc_length = 16;
  growable->used = 0;
  growable->size = unitSize;
}

void gm_push(GrowMem * growable, void * object)
{
  if (growable->used >= growable->alloc_length)
    {
      growable->alloc_length *= 2;
      growable->memory = (char *) reallocate(growable->memory,
					     growable->alloc_length
					     * growable->size);
      growable->current = (growable->memory + growable->used * growable->size);
    }
  memcpy(growable->current, object, growable->size);
  growable->current += growable->size;
  growable->used++;
}

size_t size_of(GrowMem * growable)
{
  return growable->used;
}

void lookup(GrowMem * growable, size_t item, void * object)
{
  memcpy(object, growable->memory + item * growable->size, growable->size);
}

void freeGrowMem(GrowMem * growable)
{
  free(growable->memory);
  growable->memory = 0;
  growable->current = 0;
  growable->alloc_length = 0;
  growable->used = 0;
}

void * mallocate(size_t bytes)
{
  void * mem;
  if (!bytes) bytes++;
  mem = malloc(bytes);
  if (!mem)
    {
      fprintf(stderr, "Memory full - could not allocate %ld bytes\n", bytes);
      exit(1);
    }
  return mem;
}

void * reallocate(void * old, size_t bytes)
{
  void * mem;
  if (!bytes) bytes++;
  mem = realloc(old, bytes);
  if (!mem)
    {
      fprintf(stderr, "Memory full - could not allocate %ld bytes\n", bytes);
      exit(1);
    }
  return mem;
}

void * callocate(size_t num_elts, size_t elt_size)
{
  void * mem;
  if (!num_elts) num_elts++;
  mem = calloc(num_elts, elt_size);
  if (!mem)
    {
      size_t bytes = num_elts * elt_size;
      fprintf(stderr, "Memory full - could not allocate %ld bytes\n", bytes);
      exit(1);
    }
  return mem;
}

int main(int argc, char *argv[])
{			/* ace file input */
    GapIO *dbhandle;

    char project[1024] = {0};		/* project to read */
    char version[1024] = {'0', 0};
    char acefile[1024] = {0};
    char contig_prefix[1024] = {'C', 'o', 'n', 't', 'i', 'g', 0};
    int cutoff = 100;
    int staden = 0;
    int fasta = 0;
    int caf = CAF;
    int quality = 1;
    int silent = 0;
    int pads = 0;
    int depad = 0;
    int preserve = 1;
    int bayes = 0;
    int id = 0;


    FILE * fp;
    int old_stdout_des;
    int stderr_des;
    int new_stdout_des;
    FILE * real_stdout;

    errno = 0;

    /* GAP4 libraries write to STDOUT grrr... */
    old_stdout_des = fileno(stdout);
    if (old_stdout_des < 0) {
      fprintf(stderr, "ERROR: Couldn't get file descriptor for stdout\n");
      exit(1);
    }
    stderr_des = fileno(stderr);
    if (stderr_des < 0) {
      fprintf(stderr, "ERROR: Couldn't get file descriptor for stderr\n");
      exit(1);
    }
    new_stdout_des = dup(old_stdout_des);
    if (new_stdout_des < 0) {
      fprintf(stderr, "ERROR: Couldn't dup file descriptor for stdout\n");
      exit(1);
    }
    real_stdout = fdopen(new_stdout_des, "w");
    if (!real_stdout) {
      fprintf(stderr, "ERROR: Couldn't get file pointer for real_stdout\n");
      exit(1);
    }
    if (dup2(stderr_des, old_stdout_des) < 0) {
      fprintf(stderr, "ERROR: Couldn't redirect stdout to stderr\n");
      exit(1);
    }

    (getarg("-version",version,argc,argv)) || (strcpy(version,"0"));
    (getint("-cutoff",&cutoff,argc,argv)) || (cutoff = 100);
    (getboolean("-staden",&staden,argc,argv)) || (staden = 0);
    (getboolean("-fasta",&fasta,argc,argv)) || (fasta = 0);
    (getboolean("-caf",&caf,argc,argv)) || (caf = CAF);
    (getboolean("-quality",&quality,argc,argv)) || (quality = 1);
    (getboolean("-silent",&silent,argc,argv)) || (silent = 0);
    (getboolean("-pads",&pads,argc,argv)) || (pads = 0);
    (getboolean("-depad",&depad,argc,argv)) || (depad = 0);
    (getboolean("-preserve",&preserve,argc,argv)) || (preserve = 1);
    (getboolean("-bayesian",&bayes,argc,argv)) || (bayes = 0);
    (getarg("-contigs",contig_prefix,argc,argv)) || (strcpy(contig_prefix,staden?"":"Contig"));
    (getint("-id",&id,argc,argv)) || (id = 0);


    (fp = argfile("-ace", "w", argc, argv, acefile )) || (fp=real_stdout);

    getarg("-project",project,argc,argv); /* gap project name to create */
    gethelp(argc,argv);

    if ( ! *project ) {
	usage(); exit(2);
    }

    /* Initialize various bits of gap4 */
    Tk_utils_Init(Tcl_CreateInterp());
    set_dna_lookup();

    dbhandle = open_database(project,version);

    if (depad && !pads) pads = 1;

    gap2caf(fp, dbhandle, cutoff, staden, fasta, quality, silent,
	    pads, depad, preserve, contig_prefix, bayes, id);
    
    fclose(fp);

    close_db(dbhandle);

    return 0;
} 

void usage(void)
{
  fprintf(stderr, "gap2caf $Id: newgap2caf.c,v 1.9 2006/05/10 16:12:14 rmd Exp $\n");
  fprintf(stderr, "Usage: gap2caf -project PROJECT -version VERSION [-staden] [-cutoff #] [-ace acefile]\n");
}

GapIO *open_database(char *project, char *version)
{
  GapIO *dbhandle;
  int status;
  char * s;
  char fname[1024];

  sprintf(fname, "%s.%s.aux", project, version);
  if (access(fname, R_OK) != 0) {
      for(s=project;*s;s++) if (islower((int) *s)) *s = toupper((int) *s);
      for(s=version;*s;s++) if (islower((int) *s)) *s = toupper((int) *s);
  }

  dbhandle = open_db(project, version, &status, 0, 1);

  if (!dbhandle)
    {
      fprintf(stderr, "Error opening database\n");
      exit(1);
    }

  return dbhandle;
}

void gap2caf(FILE *fp,		 /* FILE to write to */
	     GapIO *dbhandle,    /* GAP database handle */
	     int cutoff,	 /* cutoff for consensus calculation */
	     int staden,	 /* generate names for contigs a la Staden */
	     int fasta,		 /* generate fasta output, instead of CAF */
	     int quality,	 /* generate BaseQuality information */
	     int silent,	 /* run silently */
	     int pads,		 /* preserve pads in consensus (fasta ONLY) */
	     int depad,		 /* remove pads from consensus (fasta ONLY) */
	     int preserve,	 /* preserve staden read id */
	     char *contig_prefix,/* prefix for contig names (when not Staden)*/
	     int bayes,		 /* Use Bayesian consensus generation */
             int id)             /* Id of contig */
           


{
  GContigs cInfo = {0};
  int numContigs;
  int contig;
  char *cname;
  char *used_annotations, *used_notes;
  char *tc;
  int cnum;

  numContigs = NumContigs(dbhandle);
 
  used_annotations = (char *) callocate(Nannotations(dbhandle) + 1, sizeof(char));
  used_notes = (char *) callocate(Nnotes(dbhandle) + 1, sizeof(char));
  
  for (contig = 1; contig <= numContigs; contig++)
    {
      if(id){
	tc = get_contig_name(dbhandle,contig);
	cnum = get_gel_num(dbhandle,tc,GGN_ID);
	if(cnum == id) {
	  writeContig(fp, dbhandle, cutoff, staden, fasta, quality, silent, pads,
		      depad, preserve, contig_prefix, bayes,
		      contig, used_annotations, used_notes, id);
	}else{
	  continue;
	}
      }else{
	writeContig(fp, dbhandle, cutoff, staden, fasta, quality, silent, pads,
		    depad, preserve, contig_prefix, bayes,
		    contig, used_annotations, used_notes, id);
      }
    }
}

void writeContig(FILE *fp, GapIO *dbhandle, int cutoff, int staden,
		 int fasta, int quality, int silent, int pads,
		 int depad, int preserve, char *contig_prefix,
		 int bayes, int contig, char *used_annotations, 
                 char *used_notes, int id)
{
  char *cname;
  GContigs contigInfo = {0};
  GReadings readInfo = {0};
  GrowMem assembledFrom;
  GCardinal reading;
  char *consensus;
  char *rname;
  float *qual;
  float cut;
  int qualcut;

    
  if (contig_read(dbhandle, contig, contigInfo))
    {
      fprintf(stderr, "Couldn't read infomation for contig %d\n", contig);
      exit(1);
    }

  if (contigInfo.length == 0) {
    fprintf(stderr, "Ignoring contig %d as it has zero length\n", contig);
    return;
  }

  if(id) {
    fprintf(stderr, "\n *-* Writing caffile for contig #%i *-* \n",id);
  }
  
  cname = (char *) mallocate((strlen(contig_prefix) + 100) * sizeof(char));
  initGrowMem(&assembledFrom, sizeof(AFInfo));
  consensus = (char *) mallocate(contigInfo.length * sizeof(char));
  qual = (float *) mallocate(contigInfo.length * sizeof(float));

  if (staden)
    sprintf(cname, "%s%05d", contig_prefix, contigInfo.left);
  else
    sprintf(cname, "%s%d", contig_prefix, contig);
  
  if (!silent) fprintf(stderr, "Sequence : %s\n", cname);
  
  if (!fasta)
    {
      size_t i;
      
      for (reading = contigInfo.left; reading != 0; reading = readInfo.right)
        {
          /*fprintf(stderr, ">Processing reading %d...<\n", reading);*/
	  
	  if (gel_read(dbhandle, reading, readInfo))
	    {
	      fprintf(stderr,
		      "Couldn't get information for read %d\n", reading);
	      exit(1);
	    }
          
            writeRead(fp, dbhandle, quality, silent, preserve,
                      reading, &readInfo, &assembledFrom, used_annotations,
                      used_notes);
	}
          
      /* Print Contig object */
      fprintf(fp, "Sequence : %s\n", cname);
      fprintf(fp, "Is_contig\n");
      fprintf(fp, "Padded\n");
      
      for (i = 0; i < size_of(&assembledFrom); i++)
	{
          AFInfo af_line;
          lookup(&assembledFrom, i, &af_line);
	  fprintf(fp, "Assembled_from %s %ld %ld %ld %ld\n",
		  get_read_name(dbhandle, af_line.read),
		  af_line.startInContig, af_line.endInContig,
		  af_line.startInRead, af_line.endInRead);
	}

      if (writeTags(fp, dbhandle, contigInfo.annotations, used_annotations))
	fprintf(stderr, "WARNING: Loop in annotation list for contig %s\n"
		,cname);

      if (writeNotes(fp, dbhandle, contigInfo.notes, used_notes))
	fprintf(stderr, "WARNING: Loop in note list for contig %s\n"
		,cname);
      

      fprintf(fp, "\n");
    }
  
  consensus_mode = (bayes) ? CONSENSUS_MODE_CONFIDENCE : CONSENSUS_MODE_FREQ;
  
  cut = (float) cutoff / 100.0;
  qualcut = (bayes) ? 0 : -1;
  
  if (calc_consensus(contig, 1, contigInfo.length, CON_SUM,
		     consensus, NULL, qual, NULL,
		     (float) cut, qualcut, database_info, dbhandle))
    {
      fprintf(stderr, "Couldn't calculate consensus for contig %d\n", contig);
      exit(1);
    }
  
  writeDNA(fp, cname, consensus, contigInfo.length, fasta, depad, pads);

  if (bayes && quality && !fasta)
    writeConsensusQuality(fp, cname, qual, contigInfo.length);
  
  
  freeGrowMem(&assembledFrom);
  free(cname);
  free(consensus);
  free(qual);
}

void writeRead(FILE *fp, GapIO *dbhandle, int quality,
	       int silent, int preserve, GCardinal reading,
	       GReadings * readInfo, GrowMem * assembledFrom,
	       char *used_annotations, char *used_notes)
{
  char * rname;
  AFInfo AFentry;
  GCardinal readLen;
  int readLen2;
  int clipLeft;
  int clipRight;
  char * dna;
  int1 * qual;
  int2 * original_positions;


  rname = get_read_name(dbhandle, reading);
  AFentry.read = reading;
  readLen = readInfo->length;
  dna = mallocate(readLen * sizeof(char));
  qual = mallocate(readLen * sizeof(int1));
  original_positions = mallocate(readLen * sizeof(int2));
  
  if (io_read_seq(dbhandle, reading, &readLen2,
		  &clipLeft, &clipRight, dna, qual, original_positions))
    {
      fprintf(stderr, "FATAL: Couldn't read gel number %d\n", reading);
      exit(1);
    }

  if (readInfo->sense == GAP_SENSE_REVERSE)
    {
      int start = clipLeft;
      int end   = clipRight;
      io_complement_seq(&readLen2, &start, &end,
			dna, qual, original_positions);
      AFentry.startInContig = (readInfo->position
			       + abs(readInfo->sequence_length) - 1);
      AFentry.endInContig = (readInfo->position);
      AFentry.startInRead = (readInfo->length - (readInfo->end - 1) + 1);
      AFentry.endInRead = (readInfo->length - (readInfo->start + 1) + 1);
    }
  else
    {
      AFentry.startInContig = (readInfo->position);
      AFentry.endInContig = (readInfo->position
			     + abs(readInfo->sequence_length) - 1);
      AFentry.startInRead = (readInfo->start + 1);
      AFentry.endInRead = (readInfo->end - 1);
    }
  gm_push(assembledFrom, &AFentry);
  
  writeDNA(fp, rname, dna, readLen, 0, 0, 0);

  if (quality) writeQuality(fp, rname, qual, readLen);

  fprintf(fp, "Sequence : %s\n", rname);
  fprintf(fp, "Is_read\n");
  fprintf(fp, "Padded\n");
  if (preserve) fprintf(fp, "Staden_id %d\n", reading);
  
  fprintf(fp, "Clipping QUAL %ld %ld\n", AFentry.startInRead, AFentry.endInRead);
  
  writeAlignSCF(fp, rname, original_positions, readLen);
  
  if (readInfo->trace_type && readInfo->trace_name)
    writeTraceFile(fp, dbhandle, reading);
  
  writeReadInfo(fp, dbhandle, reading, readInfo, silent);
  if (writeTags(fp, dbhandle, readInfo->annotations, used_annotations))
    fprintf(stderr, "WARNING: Loop in annotation list for read %s\n", rname);

  if (writeNotes(fp, dbhandle, readInfo->notes, used_notes))
    fprintf(stderr, "WARNING: Loop in note list for contig %s\n", rname);
  fprintf(fp, "\n");

  free(dna);
  free(qual);
  free(original_positions);
}

void writeDNA(FILE *fp, char *name, char *dna, int len, int fasta, int depad,
	      int pads)
{
  int i = 0;
  int left;
  
  if (fasta)
    {
      fprintf(fp, ">%s\n", name);
      if (depad) {
	char *out = dna;
	for (i = 0; i < len; i++) {
	  if (dna[i] != '*') { *out++ = dna[i]; }
	}
	len = out - dna;
      }
      staden2fasta(dna, len, pads);
    }
  else
    {
      staden2caf(dna, len);
      fprintf(fp, "DNA : %s\n", name);
    }

  for(i=0; i<len; i+=60)
    {
      left = len - i;
      fwrite(dna + i, (left > 60) ? 60 : left, 1, fp);
      fprintf(fp, "\n");
    }

  if (!fasta)
    fprintf(fp, "\n");
}

void staden2caf(char *seq, int length)
				/* translate codes to CAF from staden */
{
  static char mx[256];
  static unsigned char bases[]  = "CTAGctag*-";
  static unsigned char cbases[] = "CTAGctag-n";
  static int init = 0;

  int i;

  if (!init)
    {
      init++;
      for (i = 0; i < 256; i++) { mx[i] = 'n'; }
      
      for (i = 0; i < sizeof(bases); i++) {
	mx[bases[i]] = cbases[i];
      }
    }

  for (i = 0; i < length; i++) seq[i] = mx[(unsigned char) seq[i]];
}

void staden2fasta(char *seq, int length, int pads)
				/* translate codes to CAF from staden */
{
  static char mx[256];
  static unsigned char bases[]  = "CTAGctag-";
  static unsigned char cbases[] = "ctagctagn";
  static int init = 0;

  int i;

  if (!init)
    {
      init++;
      for (i = 0; i < 256; i++) { mx[i] = 'n'; }
      
      for (i = 0; i < sizeof(bases); i++) {
	mx[bases[i]] = cbases[i];
      }
    }

  mx['*'] = (pads ? '*' : 'n');

  for (i = 0; i < length; i++) seq[i] = mx[(unsigned char) seq[i]];
}

void writeQuality(FILE *fp, char *name, int1 *qual, int len)
{
    int i = 0;

    fprintf(fp, "BaseQuality : %s\n", name);

    for(i = 0; i < len; i++) {
	/*
	 * NOTE
	 * XGAP sets confidence of base to 0 or 100 when an edit is made.
	 * Phrap uses for the same thing 98 and 99 instead.
	 * These are the CAF standard also.
	 * We cannot deal with the confidences being set manually low in GAP
	 * but deal with high option here.
	 */
	fprintf(fp, "%d ", qual[i] != 100 ? qual[i] : 99);
    }

    fprintf(fp,"\n");
    fprintf(fp,"\n");
}

void writeConsensusQuality(FILE *fp, char *name, float *qual, int len)
{
  int i = 0;
  
  fprintf(fp, "BaseQuality : %s\n", name);
  
  for(i = 0; i < len; i++) {
    /*
     * NOTE
     * XGAP sets confidence of base to 0 or 100 when an edit is made.
     * Phrap uses for the same thing 98 and 99 instead.
     * These are the CAF standard also.
     * We cannot deal with the confidences being set manually low in GAP
     * but deal with high option here.
     */
    int qual_val = (int) qual[i];
    fprintf(fp, "%d ", qual_val <= 99 ? qual_val : 99);
  }
  
  fprintf(fp,"\n");
  fprintf(fp,"\n");
}

void writeAlignSCF(FILE *fp, char *name, int2 *op, int len)
{
  int i = 0;      /* Where we are in the original_positions array      */
  int j = 0;      /* j = dna_start - i                                 */
  int dna_start;  /* Start of the current Align_to_SCF (DNA positions) */

  int scf_start;  /* Start of the current Align_to_SCF (SCF positions) */
  int scf_offset = 0; /* Offset to add due to wrapping round           */
  int allow_inc  = 1; /* Allow offset to increment                     */

  for(i = 0; i < len;) {
    /* do next Align_to_SCF line */
    if (op[i]) {

      if (allow_inc && op[i] < -16384) {
	scf_offset += 65536;
	allow_inc = 0;
      }

      dna_start = i;
      j = 0;
      scf_start = scf_offset + (int) op[i];

      do {
	i++;
	j++;
	if (i >= len) break;
	if (allow_inc && op[i] < -16384) {
	  scf_offset += 65536;
	  allow_inc = 0;
	}
	if (op[i] > 16384 && !allow_inc) {
	  allow_inc = 1;
	}

      }	while (scf_offset + op[i] == scf_start + j);

      fprintf(fp, "Align_to_SCF %d %d %d %d\n",
	      dna_start + 1, i, scf_start, scf_offset + op[i - 1]);
    } else {
      i++;
    }
  }
}

void writeTraceFile(FILE *fp, GapIO *dbhandle, int reading)
{
  char trace_type[256];
  char trace_name[256];
  
  if (!io_read_rd(dbhandle, reading, trace_name, 256, trace_type, 256))
    {
      if (!strcmp(trace_type,"SCF"))
	fprintf(fp, "SCF_File %s\n", trace_name);
    }
}

void writeReadInfo(FILE *fp, GapIO *dbhandle, int reading,
		   GReadings * readInfo, int silent)
{
  char clone[256];
  char cloning_vector[256];
  char subclone[256];
  char sequencing_vector[256];
  int length;
  int sizeMin;
  int sizeMax;
  int direction;
  int strand;
  int primer;
  int cloneId;
  int subcloneId;
  int cvecId;
  int svecId;

  if (get_read_info(dbhandle, reading, clone, 256, cloning_vector, 256,
		    subclone, 256, sequencing_vector, 256,
		    &length, &sizeMin, &sizeMax, &direction, &strand,
		    &primer, &cloneId, &subcloneId, &cvecId, &svecId))
    {
      fprintf(stderr, "Couldn't get clone information for read %d\n", reading);
      exit(1);
    }

  switch(primer)
    {
    case GAP_PRIMER_FORWARD:
    case GAP_PRIMER_REVERSE:
      fprintf(fp,"Primer Universal_primer\n");
      break;
    case GAP_PRIMER_CUSTFOR:
    case GAP_PRIMER_CUSTREV:
      fprintf(fp,"Primer Custom\n");
      break;
    default:
      break;
    }
  
  switch(direction)
    {
    case GAP_STRAND_FORWARD:
      fprintf(fp,"Strand Forward\n");
      break;
    case GAP_STRAND_REVERSE:
      fprintf(fp,"Strand Reverse\n");
      break;
    default:
      break;
    }
  
  if (readInfo->chemistry & GAP_CHEM_TERMINATOR)
    fprintf(fp,"Dye Dye_terminator\n");
  else
    fprintf(fp,"Dye Dye_primer\n");

  if (subcloneId)
    {
      fprintf(fp, "Insert_size %d %d\n", sizeMin, sizeMax);
      if (fixTemplate(subclone) && ! silent)
	fprintf(stderr, "WARNING: Fixed template to: %s\n", subclone);
      fprintf(fp,"Template %s\n", subclone);
    }
  if (cloneId) fprintf(fp, "Clone %s\n", clone);
  if (svecId) fprintf(fp, "Sequencing_vector \"%s\"\n", sequencing_vector);
}

/*Gap-ReadRaid bug workaround -- DMH
*
* Truncate template strings down
* to before the occurance of the
* first `.'
*
* Return 1 if this occured.*/

int fixTemplate(char *template)
{

    char *suffix_ptr;

    if (0 != (suffix_ptr = strstr(template, ".")))
      {
	*suffix_ptr = 0;
	return 1;
      }

    return 0;
}

int writeTags(FILE * fp, GapIO * dbhandle, GCardinal tag, char *used)
{
  GAnnotations tag_data;
  int from, to;

  GCardinal nanno = Nannotations(dbhandle);

  for (; tag && tag <= nanno && !used[tag]; tag = tag_data.next)
    {
      char *comment    = 0;
      char tag_type[5] = {0};
      char *method     = 0;
      char *newcomment = 0;

      used[tag] = 1;
      if (tag_read(dbhandle, tag, tag_data))
	{
	  fprintf(stderr, "Couldn't read annotation %d\n", tag);
	  exit(1);
	}
      
      type2str(tag_data.type, tag_type);

      if (tag_data.annotation)
	{
	  if ((comment = TextAllocRead(dbhandle, tag_data.annotation)) == 0)
	    {
	      fprintf(stderr,
		      "Couldn't read comment for annotation %d\n", tag);
	      exit(1);
	    }

	}
      
      mapTypeToMethod(tag_type, comment, &method, &newcomment);
      if (!newcomment) newcomment = comment;

      if (GAP_STRAND_FORWARD == tag_data.strand) {
	from = tag_data.position;
	to = tag_data.position + tag_data.length - 1;
      } else {
	from = tag_data.position + tag_data.length - 1;
	to = tag_data.position;
      }
      
      /* Some tags have special representation in CAF */
      if (0==strncmp(tag_type, "SVEC", 4)) {

	fprintf(fp,"Seq_vec %s %d %d", method, from, to);

      } else if (0==strncmp(tag_type, "CVEC", 4)) {

	fprintf(fp,"Clone_vec %s %d %d", method, from, to);

      } else if (0==strncmp(tag_type, "STOL", 4)) {

	fprintf(fp,"Stolen");

      } else {

	fprintf(fp,"Tag %s %d %d", method, from, to);

      }
      
      if (newcomment && *newcomment)
	{
	  char * s;
	  putc(' ', fp);
	  putc('"', fp);
	  for (s = newcomment; *s; s++)
	    {
	      switch(*s)
		{
		case '\n': putc('\\', fp); putc('n', fp); break;
		case '"' : putc('\\', fp); putc('"', fp); break;
		case '\\': putc('\\', fp); putc('\\', fp); break;
		case '\377':
		  fprintf(fp, "\\%03o", (unsigned char) *s);
		  break;
		default: putc(*s,fp); break;
		}
	    }
	  putc('"', fp);
	}
      fprintf(fp,"\n");
      if (newcomment != comment) free(newcomment);
      if (comment)               free(comment);
      free(method);
    }
  /* Returns not 0 if loop found */
  return tag;
}


int writeNotes(FILE * fp, GapIO * dbhandle, GCardinal note, char *used)
{
  GNotes note_data;

  GCardinal nnote = Nnotes(dbhandle);

  for (; note && note <= nnote && !used[note]; note = note_data.next)
    {
      char   *comment     = 0;
      char   note_type[5] = {0};
      char   *method      = 0;
      char   *newcomment  = 0;
      time_t ctime = 0, mtime = 0;

      used[note] = 1;
      if (note_read(dbhandle, note, note_data))
	{
	  fprintf(stderr, "Couldn't read note %d\n", note);
	  exit(1);
	}
      
      type2str(note_data.type, note_type);

      /*
       * Fetch the actual not string using the staden id held in
       * note_data.annotation If there isn't a valid id then assign
       * an empty string.
       */
      if (note_data.annotation)
	{
	  if ((comment = TextAllocRead(dbhandle, note_data.annotation)) == 0)
	    {
	      fprintf(stderr,
		      "Couldn't read comment for note %d\n", note);
	      exit(1);
	    }
	}

      mapTypeToMethod(note_type, comment, &method, &newcomment);
      if (!newcomment) newcomment = comment;
      
#if SIZEOF_TIME_T == 4
      /*
       * Error check. Gap databse contains time stamps that are too big
       * to represent in this machines time_t.
       */
      if(note_data.ctime_top != 0 || note_data.mtime_top != 0) {
          fprintf(stderr, "Gap database contains unsupported timestamps > 32 bits... exiting\n");
          exit(1);
      }
      fprintf(fp,"Note %s %d %d", method, note_data.ctime, note_data.mtime);
#elif SIZEOF_TIME_T == 8

      /*
       * NB. THIS CODE NOT CHECKED. Currently not supported by Gap, or in
       * cross platform format.
       */

      ctime = mtime = 0;

      ctime |= note_data.ctime_top;
      ctime <<= 32;
      ctime |= note_data.ctime;

      mtime |= note_data.mtime_top;
      mtime <<= 32;
      mtime |= note_data.mtime;

      fprintf(fp,"Note %s %ld %ld", method, ctime, mtime);
#else
#error Unsupported time_t size (SIZEOF_TIME_T)... Exiting
#endif

      /*
       * Output the string and escape special characters where required.
       * I'm not sure of the purpose for the ascii(255) case.
       */
      if (newcomment && *newcomment)
	{
	  register char *s;
	  putc(' ', fp);
	  putc('"', fp);
	  for (s = newcomment; *s; s++)
	    {
	      switch(*s)
		{
		case '\n': putc('\\', fp); putc('n', fp); break;
		case '"' : putc('\\', fp); putc('"', fp); break;
		case '\\': putc('\\', fp); putc('\\', fp); break;
		case '\377':
		  fprintf(fp, "\\%03o", (unsigned char) *s);
		  break;
		default: putc(*s,fp); break;
		}
	    }
	  putc('"', fp);
	}
      fprintf(fp,"\n");
      if (newcomment != comment) free(newcomment);
      if (comment)               free(comment);
      free(method);
    }
  /* Returns not 0 if loop found */
  return note;
}

void mapTypeToMethod(char t[5],
		     char *comment,
		     char **method,
		     char **newcomment)
{
  int i;
  int err;
  
  if ((err = getTagValue(comment, 
			 "Method",
			 method,
			 newcomment,
			 1 /*removeIfFound*/)) || ! *method)
    {
      /* No method */
      *method = mallocate(5 * sizeof(char));
      strncpy(*method, t, 4);
      
      /* handle tag types less than four characters */
      for(i = 3; i >= 0 && (*method)[i] == ' '; i--);
      (*method)[i + 1] = '\0';

      *newcomment = NULL;
    }
}

int getTagValue (char *comment,
		 char *tag,
		 char **value,
		 char **newcomment,
		 int removeIfFound)
{
  int i;
  int taglen;
  int end;
  int valuelen;
  char sep;
  
  if (NULL == comment || strncmp(comment,"CAF",3) || ! (sep = comment[3])) {
    *value = NULL;
    return 1;
  }
  
  taglen = strlen(tag);
  
  for(i = 4; comment[i] && comment[i] != sep; i = end) {
    
    /* check for this tag */
    if (0 == strncmp(comment + i, tag, taglen)) {
      
      if (comment[i + taglen] == '=') {
	
	/* tag-value pair found */
	end = find_end_of_value(comment, i + taglen + 1, sep);
	if ( 0 == comment[end] ) {
	  *value = NULL;	
	  return 1;
	}
	
	valuelen = end - (i + taglen + 1);
	*value = mallocate((valuelen + 1) * sizeof(char));
	strncpy(*value, comment + i + taglen + 1, valuelen);
	(*value)[valuelen] = '\0';
	
	/* remove tag-value from comment */
	if (removeIfFound)
	  excise(comment, i, taglen + valuelen + 1 + 1, newcomment, sep);
	/* 1 for =, 1 for trailing / */
	return(0);
	
      } else if (comment[i + taglen] == sep) {
	
	/* boolean tag found */
	*value = NULL;
	/* remove tag from comment */
	if (removeIfFound)
	  excise(comment, i, taglen + 1, newcomment, sep);
	/* 1 for trailing / */
	return(0);
      }
    }
    
    end = find_end_of_value(comment, i, sep) + 1;
    
    /* skip to next tag */
  }
  
  return(1);
  
}

int find_end_of_value (char *comment, int i,char sep)
{
    for(; comment[i] && comment[i] != sep; i++) {

	if (comment[i] == '\\' && comment[i+1]) i++;

    }

    return i;
}

void excise(char *comment,
	    int from,
	    int len,
	    char **newcomment,
	    char sep)
{
  int commentlen = strlen(comment);
  *newcomment = mallocate((commentlen - len + 1) * sizeof(char));
  sprintf(*newcomment, "%.*s%s", from, comment, comment + from + len);
  
  /* remove magic if no tags left */
  /* assert(strlen(*newcomment) >= 5) */
  if ((*newcomment)[0] == 'C' &&
      (*newcomment)[1] == 'A' &&
      (*newcomment)[2] == 'F' &&
      (*newcomment)[3] == sep &&
      (*newcomment)[4] == sep) {
    int i;
    for (i = 0; 0 != ((*newcomment)[i] = (*newcomment)[i + 5]); i++) {}
  }
}
